/* ----------------------------------------------------------------------
   LAMMPS - Large-scale Atomic/Molecular Massively Parallel Simulator
   http://lammps.sandia.gov, Sandia National Laboratories
   Steve Plimpton, sjplimp@sandia.gov

   Copyright (2003) Sandia Corporation.  Under the terms of Contract
   DE-AC04-94AL85000 with Sandia Corporation, the U.S. Government retains
   certain rights in this software.  This software is distributed under 
   the GNU General Public License.

   See the README file in the top-level LAMMPS directory.
------------------------------------------------------------------------- */

#include "math.h"
#include "mpi.h"
#include "string.h"
#include "stdlib.h"
#include "fix_redox.h"
#include "atom.h"
#include "random_mars.h"
#include "update.h"
#include "error.h"

using namespace LAMMPS_NS;

#define BIG 1.0e10

/* ---------------------------------------------------------------------- */

FixRedOx::FixRedOx(LAMMPS *lmp, int narg, char **arg) :
  Fix(lmp, narg, arg)
{
  if (narg < 14) error->all("Illegal fix redox command");

  MPI_Comm_rank(world,&me);

  nevery = atoi(arg[3]);
  red_type = atoi(arg[4]);
  ox_plane = atof(arg[5]);
  ox_type = atoi(arg[6]);
  red_plane = atof(arg[7]);
  dz = atof(arg[8]);
  ne = atoi(arg[9]);
  alpha = atof(arg[10]);
  enot = atof(arg[11]);
  beta = atof(arg[12]);
  knot = atof(arg[13]);
  tau = atof(arg[14]);
  int seed = atoi(arg[15]);

  force_reneighbor = 1;
  next_reneighbor = -1;

  scalar_flag = 1;
  global_freq = nevery;
  extscalar = 0;
  random = new RanMars(lmp,seed + me);

  // initialize array sizes, need a better way to do this 

  e_exchange = 0.0;
}

/* ---------------------------------------------------------------------- */

FixRedOx::~FixRedOx()
{
  delete random;
}

/* ---------------------------------------------------------------------- */

int FixRedOx::setmask()
{
  int mask = 0;
  mask |= POST_INTEGRATE;
  return mask;
}

/* ---------------------------------------------------------------------- */

void FixRedOx::post_integrate()
{
  int i;
  double coord,kr,kox,kred,prob,ran;

  // only consider atoms within dz of slab 
  // map atoms back into periodic box if necessary
  // insert = location in list to insert new atom

  double **x = atom->x;
  double *q = atom->q;
  int *type = atom->type;
  int *mask = atom->mask;
  int nlocal = atom->nlocal;
  int *tag = atom->tag;
  int react =0;
  double eswap = 0.0;

  for (i = 0; i < nlocal; i++) {
    if (mask[i] & groupbit) {
      coord = x[i][2];
      if (type[i] == red_type){ 
        if (coord >= ox_plane && coord <= ox_plane+dz) {
          kr=knot*exp(-beta*fabs(ox_plane-coord));
          kred=kr*exp((1.0-alpha)*ne*38.82*enot);
          prob=1-exp(-kred*tau);
          ran=random->uniform();
          if (prob >= ran){
   fprintf(screen,"ox p=%f mc=%f type=%d tag=%d z=%f \n",prob,ran,type[i],tag[i],coord = x[i][2]);
            react = 1;
            type[i]=ox_type;
            q[i]=1.0;
            eswap += 1.0;
          }
	}
      }
      if (type[i] == ox_type){
        if (coord <= red_plane && coord >= red_plane-dz) {
          kr=knot*exp(-beta*fabs(red_plane-coord));
          kox=kr*exp((1.0-alpha)*ne*38.82*enot);
          prob=1-exp(-kox*tau);
          ran=random->uniform();
          if (prob >= ran){
   fprintf(screen,"red p=%f mc=%f type=%d tag=%d z=%f \n",prob,ran,type[i],tag[i],coord = x[i][2]);
            react = 1;
            type[i]=red_type;
            q[i]=0.0;
            eswap += 1.0;
          }
        } 
      }
    }
  }  

//  reneighbor if ther is a reaction

  int react_all;
  MPI_Allreduce(&react,&react_all,1,MPI_INT,MPI_SUM,world);
  if(react_all) next_reneighbor = update->ntimestep;

  // tally momentum exchange from all swaps

  double eswap_all;
  MPI_Allreduce(&eswap,&eswap_all,1,MPI_DOUBLE,MPI_SUM,world);
  e_exchange += eswap_all;
}

/* ---------------------------------------------------------------------- */

double FixRedOx::compute_scalar()
{
  return e_exchange;
}
